"""Insights connector module.

This module defines the InsightsConnector class for managing data processing and submission to an Insights server.
"""

import itertools
import logging
import os
import math
import traceback

import ormsgpack
import polars as pl
import queue
import requests as rq
import threading
import time


from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime, timezone
from io import BytesIO
from typing import Optional, Any, TypeVar

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item

from dataspree.inference.insights.encoding import Encoding, EncoderManager
from dataspree.inference.insights.schemata import BaseSchema
from dataspree.inference.insights.tables import TableCollection, Table
from dataspree.inference.insights.exceptions import WrongCredentialsError, UnknownAuthenticationError

logger = logging.getLogger(__name__)

TTable = TypeVar('TTable', bound=Table[BaseSchema])


class InsightsConnector:
    """Manages the connection to an Insights server for data batching and submission.

    This class handles authentication, batching, and submission of data to an Insights server. It provides
    methods to register tables, enqueue data, and manage the lifecycle of background threads for processing
    and submission.
    """

    def __init__(
        self,
        source_id: int = 0,
        files: dict[str, bool] | None = None,
        server_url: str = 'http://localhost:6726',
        batch_sleep_time: float = 0.1,
        username: str = 'admin',
        password: str = 'admin',
        encoding: Encoding = Encoding.PNG,
        persisted: bool = True,
        max_batch_size: int = 15,
        max_buffer_length: int = 100,
        num_encoding_workers: int = 3,
    ) -> None:
        """Initializes the InsightsConnector.

        Args:
            source_id (int): The ID of the data source.

            files (dict[str, bool] | None): the name of the attached files and whether they are a new file.

            server_url (str): The URL of the Insights server.

            batch_sleep_time (float): Time interval in seconds between batch processing.

            username (str): Username for authentication.

            password (str): Password for authentication.

            encoding (Encoding): the used encoding.

            persisted (bool): Whether the values are stored in insights, or just visualized temporarily.

            max_buffer_length (int): Max number of items that are buffered for the encoding.
                                     Smaller or equal to 0 to never drop.

            max_batch_size (int): Max number of frames per request.
                                  Smaller or equal to 0 to never drop.

            num_encoding_workers (int): Number of workers to spawn for encoding.
        """
        self.source_id: int = source_id
        self.files: dict[str, bool] = files or {}
        self.encoding: Encoding = encoding
        self.persisted: bool = persisted
        self.server_url: str = server_url
        self.batch_sleep_time: float = batch_sleep_time
        self.username: str = username
        self.password: str = password
        self.max_batch_size: int = max_batch_size
        self.max_buffer_length: int = max_buffer_length
        self.num_encoding_workers: int = num_encoding_workers

        self.tables = TableCollection()

        self.http: Optional[rq.sessions.Session] = None
        self.headers: Optional[dict[str, str]] = None
        self.__running: bool = False

        self.buffer_lock: threading.Lock = threading.Lock()
        self.buffer: list[Item] = []

        self.payload_queue: queue.Queue[tuple[bytes, int]] = queue.Queue(maxsize=300)

        self.batch_thread: Optional[threading.Thread] = None
        self.submit_thread: Optional[threading.Thread] = None

        self._dropped_items: int = 0
        self._accepted_items: int = 0
        self._encoded_items: int = 0
        self._sent_items: int = 0

        self.encoding_manager: Optional[EncoderManager] = None

    @property
    def insights_connector_running(self) -> bool:
        """Return whether Insights is running or currently in the process of being stopped."""
        return self.__running or self.submit_thread is not None or self.batch_thread is not None

    def status_update_string(self) -> str:
        """Return status update."""
        return (
            f'insights connector with source id {self.source_id}: rejected / accepted / encoded / '
            f'sent: {self._dropped_items}, {self._accepted_items},  {self._encoded_items},  {self._sent_items}.'
        )

    def register_table(self, table: type[TTable]) -> TTable:
        """Registers a new table with the connector.

        Args:
            table (type[Table[BaseSchema]]): The table class to register.

        Returns:
            Table[BaseSchema]: The registered table instance.
        """
        return self.tables.register_table(table)

    def register_content(self, item: Item) -> None:
        """Adds an item to the buffer for batch processing.

        Args:
            item (Item): The item to add to the buffer.

        Raises:
            ValueError: If the connector is not running.
        """
        if not self.__running:
            raise ValueError('Insights connector needs to be started in order to accept data.')

        with self.buffer_lock:
            if 0 < self.max_buffer_length <= len(self.buffer):
                self._dropped_items += 1
                logger.error(
                    f'Dropping value. Exceeded max buffer size. '
                    f'Rejected / Accepted / encoded / sent: {self._dropped_items}'
                    f', {self._accepted_items},  {self._encoded_items},  {self._sent_items}.'
                )
            else:
                self._accepted_items += 1
                self.buffer.append(item)

    def initialize(self):
        self.http = rq.Session()
        self.cookie: str = self._authenticate(
            http=self.http, username=self.username, password=self.password, server_url=self.server_url
        )
        self.headers = {'Content-Type': 'application/msgpack', 'Cookie': f'id={self.cookie}'}
        self.encoding_manager = EncoderManager(self.encoding, 3, self.num_encoding_workers)

    def start(self) -> None:
        """Starts the connector, enabling data collection and processing.

        Raises:
            ValueError: If the connector is already running.
        """
        if self.insights_connector_running:
            raise ValueError('Connector is already running (or currently being stopped). Call stop first!')

        self.__running = True

        self.initialize()

        self.batch_thread = threading.Thread(target=self._process_batch, daemon=True)
        self.batch_thread.start()

        self.submit_thread = threading.Thread(target=self._submit_tasks, daemon=True)
        self.submit_thread.start()

    def stop(self) -> None:
        """Stops the connector and gracefully shuts down threads.

        Ensures that the batch thread is stopped before the submit thread to
        deliver any remaining messages and not cause a deadlock.
        """
        logger.info('Stopping Insights Connector!')
        self.__running = False

        if (bt := self.batch_thread) is not None:
            logger.info('Joining batch thread')
            bt.join()
            self.batch_thread = None

        if (st := self.submit_thread) is not None:
            logger.info('Joining submit thread')
            st.join()
            self.submit_thread = None

        if (em := self.encoding_manager) is not None:
            logger.info('Stopping encoding manager.')
            em.stop()

        logger.info(f'Stopped. {self.status_update_string()}')

    @staticmethod
    def _authenticate(http: rq.sessions.Session, server_url: str, username: str, password: str) -> str:
        """Authenticates with the server and retrieves the session cookie.

        Returns:
            str: The session cookie.

        Raises:
            ValueError: If authentication fails.
        """
        try:
            msg_json = {'username': os.path.expandvars(username), 'password': os.path.expandvars(password)}
            response = http.post(
                f'{os.path.expandvars(server_url)}/api/auth/login',
                json=msg_json,
                timeout=5,
            )
            if response.status_code == 401:
                raise WrongCredentialsError('Invalid credentials provided.')

            if response.status_code != 200:
                raise UnknownAuthenticationError(f'Authentication failed with status code {response.status_code}.')

            logger.info('Successfully authenticated with the Insights server.')
            session_cookie = response.cookies.get('id')
            if not session_cookie:
                raise UnknownAuthenticationError('Session cookie is missing in the response.')

            return session_cookie

        except rq.RequestException as e:
            raise UnknownAuthenticationError(f'Authentication request failed: {e}')

    def _process_batch(self) -> None:
        """Background thread that processes batches at regular intervals.

        Collects buffered items, processes them into data frames, and enqueues
        the resulting payloads for submission.
        """
        try:
            while True:
                time.sleep(self.batch_sleep_time)

                with self.buffer_lock:
                    items, self.buffer = self.buffer, []

                    if not items and not self.__running:
                        break

                if not len(items):
                    time.sleep(0.1)

                for i in range(math.ceil(len(items) / self.max_batch_size)):
                    current_items = items[i * self.max_batch_size : (i + 1) * self.max_batch_size]
                    for item in current_items:
                        self.tables.from_item(item)

                    try:
                        if data := self.tables.data():
                            payload = self.encode_data(data, current_items)
                            try:
                                self.payload_queue.put((payload, len(current_items)), timeout=1)
                                logger.debug('Payload enqueued for sending.')
                            except queue.Full:
                                logger.error(
                                    f'Send queue ful. Dropping {len(items)} encoded items. '
                                    f'Rejected / Accepted / encoded / sent: {self._dropped_items}'
                                    f', {self._accepted_items},  {self._encoded_items},  {self._sent_items}.'
                                )

                    finally:
                        self.tables.clear()

        except BaseException as e:
            logger.error(f'An exception occurred in the batch processing task: {e}')

            if is_application_debug():
                traceback.print_exc()

            raise e from None

        finally:
            logger.info('Batch processing thread terminated!')

    def encode_item(self, item: Item) -> bytes | None:
        """Encode data from a single item."""
        return self.encode_items([item])

    def encode_items(self, items: list[Item]) -> bytes | None:
        """Encode data from a list of items."""
        for item in items:
            self.tables.from_item(item)

        if data := self.tables.data():
            encoded_data = self.encode_data(data, items)
            self.tables.clear()
            return encoded_data

        return None

    def encode_data(self, data_dicts: dict[str, dict[str, list[Any]]], items: list[Item]) -> bytes:
        """Builds data frames from data dictionaries.

        Args:
            data_dicts (dict[str, dict[str, list[Any]]]): A dictionary of table names
                mapped to their data content.

            items (list[Item]): original list of items.

        Return:
            (bytes) the encoded data.
        """
        data_frames: dict[str, bytes] = {}
        for table_name, table_data in data_dicts.items():
            table = self.tables[table_name]
            bytes_io = BytesIO()
            try:
                df = pl.DataFrame(table_data, schema=table.polars_schema())
                if df.is_empty():
                    logger.debug(f"Skipping Parquet encoding for empty table '{table_name}'.")
                    continue
            except TypeError as e:
                logger.fatal(f'An error occurred trying to encode table data for table {table_name}: {e}')
                if is_application_debug():
                    traceback.print_exc()

                continue

            df.write_parquet(bytes_io, compression='uncompressed')
            bytes_io.seek(0)
            data_frames[table_name] = bytes_io.read()

        batch_ts = (
            ts[-1] if (ts := data_dicts.get('frame', {}).get('timestamp')) else datetime.now(timezone.utc)
        ).isoformat()

        binary: list[tuple[str, str, list[tuple[bool, bytes]]]] = []
        if self.files:
            if (em := self.encoding_manager) is None:
                raise ValueError('Encoding manager is stopped. Cannot build payload.')

            binary = [
                (m, self.encoding.value.split(' ')[0], em.encode_images([i.get(m) for i in items], b))
                for m, b in self.files.items()
            ]

        payload_entries: list[dict[str, Any]] = [
            {
                'schema_name': table,
                'source_id': self.source_id,
                'timestamp': batch_ts,
                'data_frame': data_frames[table],
                'binary_data_columns': binary if table == 'frame' else None,
                'persist': self.persisted,
            }
            for table in data_frames
        ]

        q = ormsgpack.packb(payload_entries)
        self._encoded_items += len(items)
        return q

    def _submit_tasks(self) -> None:
        """Background thread that submits payloads to the server.

        Uses a thread pool to send payloads while handling retries and server errors.
        """
        try:
            with ThreadPoolExecutor(max_workers=2) as executor:
                futures = []
                while self.__running or not self.payload_queue.empty():
                    try:
                        futures.append(executor.submit(self.send_payload, *self.payload_queue.get(timeout=1)))
                        self.payload_queue.task_done()
                    except queue.Empty:
                        time.sleep(0.1)

                for _future in as_completed(futures):
                    pass

        except BaseException as e:
            logger.error(f'An exception occurred in the submit task: {e}')

            if is_application_debug():
                traceback.print_exc()

            raise e from None

        finally:
            logger.info('Submit thread terminated.')

    def send_payload(
        self, payload: bytes, num_items: int, max_retries: Optional[int] = 3, backoff_factor: float = 0.5
    ) -> None:
        """Sends a payload to the server with retries.

        Args:
            payload (bytes): The payload data to send.
            num_items (int): The number of items that created this payload.
            max_retries (Optional[int]): The maximum number of retry attempts. Defaults to 3.
            backoff_factor (float): The backoff factor for exponential retry delays. Defaults to 0.5.
        """
        for attempt in itertools.count(1):
            if max_retries is not None and attempt > max_retries:
                logger.error(f'Failed to send payload after {attempt - 1} attempts.')

                logger.error(
                    f'Failed to send payload after {attempt - 1} attempts. Dropping {num_items} encoded items.'
                    f'rejected / Accepted / encoded / sent: {self._dropped_items}'
                    f', {self._accepted_items},  {self._encoded_items},  {self._sent_items}.'
                )
                return

            try:
                response = self.http.post(
                    f'{self.server_url}/api/data/insert',
                    data=payload,
                    headers=self.headers,
                    timeout=10,
                )
                if response.status_code == 401:
                    logger.warning('Unauthorized request. Attempting to re-authenticate.')
                    try:
                        self.cookie = self._authenticate(
                            http=self.http, username=self.username, password=self.password, server_url=self.server_url
                        )
                        self.headers['Cookie'] = f'id={self.cookie}'
                        logger.info('Re-authentication successful. Retrying the payload.')
                        continue  # retry with the new session
                    except WrongCredentialsError:
                        logger.error('Authentication failed due to invalid credentials. Aborting retries.')
                        return  # stop retries as credentials are incorrect
                    except UnknownAuthenticationError as e:
                        logger.warning(f'Re-authentication failed with transient error: {e}. Retrying...')

                # raise an error for other non-success responses, otherwise log success and break out of loop
                response.raise_for_status()
                logger.debug('Payload sent successfully.')
                self._sent_items += num_items

                return

            except rq.RequestException as e:
                wait_time = backoff_factor * (2 ** (attempt - 1))
                logger.warning(
                    f'Could not insert data on attempt {attempt}/{max_retries or ""}. Reattempting in {wait_time:.1f}s.'
                    f'{e}'
                )
                time.sleep(wait_time)
