"""Python module containing the image encoding functionality for insights."""

import signal
import enum
from functools import partial
from typing import Callable, Optional, Any
from multiprocessing import get_context, current_process
from multiprocessing.pool import Pool as PoolType

import cv2
import time
import numpy as np
import math
from logging import getLogger

logger = getLogger(__name__)

TURBO_JPEG_FOUND: bool
try:
    from turbojpeg import TurboJPEG, TJFLAG_PROGRESSIVE, TJSAMP_GRAY

    TURBO_JPEG_FOUND = True
except (ImportError, RuntimeError) as ex:
    jpeg, TurboJPEG, TJFLAG_PROGRESSIVE, TJSAMP_GRAY = None, None, None, None
    TURBO_JPEG_FOUND = False
    logger.warning(f'Could not import turbo jpeg: {ex}.')

IMAGE_CODECS_FOUND: bool
try:
    import imagecodecs

    IMAGE_CODECS_FOUND = True
except (ImportError, RuntimeError) as ex:
    imagecodecs = None
    IMAGE_CODECS_FOUND = False
    logger.warning(f'Could not import image codecs jpeg: {ex}.')


class Encoding(str, enum.Enum):
    """Supported encoding options."""

    PNG = 'png'
    JPEG_090 = 'jpeg quality 90'
    JPEG_095 = 'jpeg quality 95'
    JPEG_100 = 'jpeg quality 100'


_turbo_jpeg: Optional[TurboJPEG]

try:
    _turbo_jpeg = TurboJPEG() if TURBO_JPEG_FOUND else None

except RuntimeError as ex:
    _turbo_jpeg = None
    TURBO_JPEG_FOUND = False
    logger.warning(f'Could not import turbo jpeg: {ex}.')


def _init_turbo_worker() -> None:
    global _turbo_jpeg

    if TurboJPEG is None:
        raise ImportError('Turbo JPEG cannot be executed because it could not be loaded.')
    _turbo_jpeg = TurboJPEG()
    logger.info(f'[{current_process().name}] initialized TurboJPEG')


def _jpeg_quality(arr: np.ndarray[Any, Any], quality: int) -> bytes:
    if _turbo_jpeg is None:
        raise ImportError('Turbo JPEG cannot be executed because it could not be loaded.')

    encoded_bytes: bytes
    if len(arr.shape) == 2 or (len(arr.shape) == 3 and arr.shape[-1] == 1):
        # Grayscale image
        if len(arr.shape) == 2:
            arr = np.repeat(arr[..., True], 3, axis=2)
        if len(arr.shape) == 3 and arr.shape[-1] == 1:
            arr = np.repeat(arr, 3, axis=2)

        encoded_bytes: bytes = _turbo_jpeg.encode(arr, jpeg_subsample=TJSAMP_GRAY)
    else:
        # Color image
        encoded_bytes: bytes = _turbo_jpeg.encode(_cvt(arr), quality=quality, flags=TJFLAG_PROGRESSIVE)
    return encoded_bytes


def _jpeg_90(arr: np.ndarray[Any, Any]) -> bytes:
    return _jpeg_quality(arr, 90)


def _jpeg_95(arr: np.ndarray[Any, Any]) -> bytes:
    return _jpeg_quality(arr, 95)


def _jpeg_100(arr: np.ndarray[Any, Any]) -> bytes:
    return _jpeg_quality(arr, 100)


def _jpeg_ls(arr: np.ndarray[Any, Any]) -> bytes:
    encoded_bytes: bytes = imagecodecs.jpegls_encode(_cvt(arr))
    return encoded_bytes


def _im_png0(arr: np.ndarray[Any, Any]) -> bytes:
    encoded_bytes: bytes = imagecodecs.png_encode(_cvt(arr), level=0)
    return encoded_bytes


def _im_png4(arr: np.ndarray[Any, Any]) -> bytes:
    encoded_bytes: bytes = imagecodecs.png_encode(_cvt(arr), level=4)
    return encoded_bytes


def _cv_png(arr: np.ndarray[Any, Any]) -> bytes:
    success, encoded = cv2.imencode('.png', _cvt(arr))
    if not success:
        raise RuntimeError('PNG encoding failed')
    encoded_bytes: bytes = encoded.tobytes()
    return encoded_bytes


def _webp4(arr: np.ndarray[Any, Any]) -> bytes:
    encoded_bytes: bytes = imagecodecs.webp_encode(_cvt(arr), level=4, lossless=True)
    return encoded_bytes


def _webp0(arr: np.ndarray[Any, Any]) -> bytes:
    encoded_bytes: bytes = imagecodecs.webp_encode(_cvt(arr), level=0, lossless=True)
    return encoded_bytes


def _cvt(arr: np.ndarray[Any, Any]) -> np.ndarray[Any, Any]:
    if len(arr.shape) == 3 and arr.shape[-1] == 3:
        return cv2.cvtColor(arr, cv2.COLOR_BGR2RGB)
    return arr


def _encode_measure(
    images: list[Optional[np.ndarray[Any, Any]]],
) -> tuple[list[Optional[bytes]], dict[str, tuple[float, float, int]]]:
    fs: list[Callable[[np.ndarray[Any, Any]], bytes]] = [
        _jpeg_90,
        _jpeg_95,
        _jpeg_100,
        _im_png0,
        _im_png4,
        _cv_png,
        _webp4,
        _webp0,
    ]
    encoded: list[Optional[bytes]] = []
    result = {}
    for fun in fs:
        t = time.perf_counter()
        encoded = [arr if arr is None else fun(arr) for arr in images]
        tt = time.perf_counter() - t
        result[fun.__name__] = (tt, sum([0.0 if r is None else len(r) for r in encoded]), len(encoded))
    return encoded, result


def _encode(images: list[Optional[np.ndarray[Any, Any]]], encoding: Encoding) -> list[Optional[bytes]]:
    function = _encoding_map[encoding][2]
    return [arr if arr is None else function(arr) for arr in images]


class EncoderManager:
    """Stores persistent variables for an encoding configuration."""

    def __init__(self, encoding: Encoding, stride: int, num_workers: int) -> None:
        """Initialize the pool."""
        self.encoding: Encoding = encoding
        self.stride: int = stride
        self.num_workers: int = max(num_workers, 1)

        original_sigint_handler = None
        try:
            original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
            ctx = get_context('spawn')
            init_function = _encoding_map[encoding][1]
            self._pool: Optional[PoolType] = ctx.Pool(processes=self.num_workers, initializer=init_function)
        finally:
            if original_sigint_handler is not None:
                signal.signal(signal.SIGINT, original_sigint_handler)

    def stop(self) -> None:
        """Destroy the pool."""
        if (pool := self._pool) is None:
            raise ValueError('Calling encode on stopped encoding manager!')
        pool.close()
        pool.join()

    def _chunks(self, images: list[Optional[np.ndarray[Any, Any]]]) -> list[list[Optional[np.ndarray[Any, Any]]]]:
        return [images[i * self.stride : (i + 1) * self.stride] for i in range(math.ceil(len(images) / self.stride))]

    def _encode_images_measure(
        self, images: list[Optional[np.ndarray[Any, Any]]], new_file: bool
    ) -> list[tuple[bool, Optional[bytes]]]:
        all_encoded: list[Optional[bytes]] = []
        results: dict[str, tuple[float, float, int]] = {}

        if self.num_workers > 1:
            if (pool := self._pool) is None:
                raise ValueError('Calling encode on stopped encoding manager!')

            rvv = pool.map(_encode_measure, self._chunks(images))
        else:
            rvv = [_encode_measure(images)]

        for encoded, rs in rvv:
            all_encoded.extend(encoded)
            for k, (t, s, n) in rs.items():
                rt, size, num = results.get(k, (0.0, 0.0, 0))
                results[k] = (rt + t, size + s, num + n)

        for k, (t, s, n) in results.items():
            print(f'{k}: {t / n}, {s / n}, {n}')

        return [(new_file, data) for data in all_encoded]

    def _encode_images(
        self, images: list[Optional[np.ndarray[Any, Any]]], new_file: bool
    ) -> list[tuple[bool, Optional[bytes]]]:
        if self.num_workers > 1:
            if (pool := self._pool) is None:
                raise ValueError('Calling encode on stopped encoding manager!')
            encoded = pool.map(partial(_encode, encoding=self.encoding), self._chunks(images))
        else:
            encoded = [_encode(images, encoding=self.encoding)]

        return [(new_file, data) for chunk in encoded for data in chunk]

    def encode_images(
        self, images: list[Optional[np.ndarray[Any, Any]]], new_file: bool
    ) -> list[tuple[bool, Optional[bytes]]]:
        """Encode the images."""
        return self._encode_images(images=images, new_file=new_file)


def available_encodings() -> list[Encoding]:
    """Return the list of all available encodings."""
    valid_encodings: list[Encoding] = [Encoding.PNG]
    if TURBO_JPEG_FOUND:
        valid_encodings.extend([Encoding.JPEG_090, Encoding.JPEG_095, Encoding.JPEG_100])
    return valid_encodings


_InitFunctionType = Callable[[], None]
_FunctionType = Callable[[np.ndarray[Any, Any]], bytes]

_encoding_map: dict[Encoding, tuple[str, Optional[_InitFunctionType], _FunctionType]] = {
    Encoding.PNG: ('png', None, _cv_png),
    Encoding.JPEG_090: ('jpeg', _init_turbo_worker, _jpeg_90),
    Encoding.JPEG_095: ('jpeg', _init_turbo_worker, _jpeg_95),
    Encoding.JPEG_100: ('jpeg', _init_turbo_worker, _jpeg_100),
}


def get_encoding_map() -> dict[Encoding, tuple[str, Optional[_InitFunctionType], _FunctionType]]:
    """Return a mapping from the available encoding options to the encoding name and function."""
    return {e: _encoding_map[e] for e in available_encodings()}
