"""Contains the default schemata used for insights."""

from enum import IntEnum

import polars as pl

from dataspree.inference.insights.schemata import BaseSchema


class SchemaFrame(BaseSchema):
    """Schema of a frame.

    Extends:
        BaseSchema

    Attributes:
        identifier (pl.String): A string identifier for the frame.
    """

    identifier: pl.String


class GeometryType(IntEnum):
    TAG = 1
    POINT = 2
    CIRCLE = 3
    SPHERE = 4
    BOX2D = 5
    BOX3D = 6
    POLYGON = 7
    SKELETON = 8
    ROI2D = 9
    ROI3D = 10


class SchemaPrediction(BaseSchema):
    """Schema of a detection.

    Extends:
        BaseSchema

    Attributes:
        frame_id (pl.Binary): Identifier for the frame associated with the detection.
        file (pl.String): File type or format code.
        model_id (pl.UInt32): Identifier for the model associated with the detection.
        confidence (pl.Float32): Confidence score for the detection.
        label (pl.UInt16): Numeric label representing the detected object class.
        class_confidences (list[pl.Float32]): List of confidence scores for each class.
        geometry_type (pl.Uint8): Type of the geometry; determines the structure of the geometry.
        geometry (list[pl.Float32]): Geometry content.
    """

    frame_id: pl.Binary
    file: pl.String
    model_id: pl.UInt32
    confidence: pl.Float32
    label: pl.UInt16
    class_confidences: list[pl.Float32]
    geometry_type: pl.UInt8
    geometry: list[pl.Float32]


class SchemaTrackedObject(BaseSchema):
    """Schema of a tracked object.

    Extends:
        BaseSchema

    Attributes:
        x (pl.Float32): X-coordinate of the tracked object.
        y (pl.Float32): Y-coordinate of the tracked object.
        z (pl.Float32): Z-coordinate of the tracked object.
        v_x (pl.Float32): X-component of the object's velocity.
        v_y (pl.Float32): Y-component of the object's velocity.
        v_z (pl.Float32): Z-component of the object's velocity.
    """

    x: pl.Float32
    y: pl.Float32
    z: pl.Float32
    v_x: pl.Float32
    v_y: pl.Float32
    v_z: pl.Float32

    # TODO: tracked_object_id: pl.Binary


class SchemaTrackedObjectDistinct(BaseSchema):
    """Schema used for the start of a tracked object to help computation.

    Extends:
        BaseSchema
    """


class SchemaEvent(BaseSchema):
    """Schema used for events.

    Extends:
        BaseSchema

    Attributes:
        type (pl.String): The type of the event.
        start_time (pl.Float32): Start time of the event.
        referenced_entity_id (pl.Binary): Identifier of the entity referenced by the event.
        referenced_table (pl.String): Name of the table containing the referenced entity.
    """

    type: pl.String
    start_time: pl.Datetime
    referenced_entity_id: pl.Binary
    referenced_table: pl.String


class SchemaEventDistinct(BaseSchema):
    """Schema used for the start of an event to help computation.

    Extends:
        BaseSchema

    Attributes:
        type (pl.String): The type of the event.
    """

    type: pl.String
