"""Contains the default tables for schemata used for insights."""

import time
import uuid

from datetime import datetime
from logging import getLogger
from typing import TypeVar, Any, Iterator

from dataspree.inference.core.item import Item

from dataspree.inference.insights.schemata.default import (
    BaseSchema,
    SchemaFrame,
    SchemaPrediction,
    SchemaTrackedObjectDistinct,
    SchemaTrackedObject,
    SchemaEventDistinct,
    SchemaEvent,
)
from dataspree.inference.insights.tables import Table

from dataspree.inference.core.regions_of_interest import roi_pos, get_region_of_interest
from dataspree.inference.insights.schemata.default import GeometryType

T = TypeVar('T', bound=BaseSchema)

logger = getLogger(__name__)


def _get(f: Any, attr: str) -> Any:
    if isinstance(f, dict):
        return f[attr]
    return getattr(f, attr)


class Frame(Table[SchemaFrame]):
    """Frame table."""

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        self.append_entry(
            id=uuid.uuid4().bytes,
            timestamp=datetime.utcfromtimestamp(item['timestamp'] * 1e-3),
            identifier=str(item['id']),
        )


class Prediction(Table[SchemaPrediction]):
    """Table for detection objects."""

    def __init__(self, geometry_types: list[GeometryType] | None = None) -> None:
        super().__init__()
        self.geometry_types: set[GeometryType] = set(
            geometry_types if geometry_types is not None else [a for a in GeometryType]
        )

    @classmethod
    def parents(cls) -> list[type[Table[T]]]:
        """Return the list of tables rows of this table depend on."""
        return [Frame]

    @classmethod
    def table_name(cls) -> str:
        """Return the name of the table."""
        return 'prediction'

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        frame_table = new_data[Frame.table_name()]
        frame_id, timestamp = next(frame_table.item_data('id')), next(frame_table.item_data('timestamp'))
        default_model_id: int = item.get(['inference', 'ai_platform_model_id'], None)
        default_file: str = str(item.get('file', '') or '')

        if GeometryType.TAG in self.geometry_types:
            self.append_classification(item, default_file, default_model_id, timestamp, frame_id)

        if GeometryType.ROI2D in self.geometry_types:
            self.append_roi_classification(item, default_file, default_model_id, timestamp, frame_id)

        b2d, b3d = GeometryType.BOX2D in self.geometry_types, GeometryType.BOX3D in self.geometry_types
        if b2d or b3d:
            self.append_detection(item, default_file, default_model_id, timestamp, frame_id, b2d=b2d, b3d=b3d)

    def append_detection(
        self, item: Item, def_fl: str, def_mdl: int, ts: datetime, fid: uuid.UUID, b2d: bool, b3d: bool
    ) -> None:
        for d in (d for d in item.get(['inference', 'detection'], []) if self.filter(d)):
            did, cf, lbl = self.get_uuid(d['id']).bytes, d['confidence'], d['label']
            model, file = d.get('model_id', def_mdl), d.get('file', def_fl)
            cs = list(d['class_confidences'].values())

            if b3d:
                try:
                    ((x, y, z), (roll, pitch, yaw)), (w, h, d) = d['pose3d'], d['size3d']
                except KeyError:
                    continue
                except ValueError as e:
                    logger.warning(f'Error extracting 3d Information: {e}')
                    continue

                self.append_entry(
                    id=did,
                    timestamp=ts,
                    frame_id=fid,
                    file=file,
                    model_id=model,
                    confidence=cf,
                    label=lbl,
                    class_confidences=cs,
                    geometry_type=GeometryType.BOX3D.value,
                    geometry=[float(f) for f in (x, y, z, w, h, d, roll, pitch, yaw)],
                )

            if b2d:
                geo_tpl = [float(f) for f in (d['x'], d['y'], d['width'], d['height'], d.get('orientation', 0.0))]
                self.append_entry(
                    id=did,
                    timestamp=ts,
                    frame_id=fid,
                    file=file,
                    model_id=model,
                    confidence=cf,
                    label=lbl,
                    class_confidences=cs,
                    geometry_type=GeometryType.BOX2D.value,
                    geometry=geo_tpl,
                )

    def append_classification(self, item: Item, def_fl: str, def_mdl: int, ts: datetime, fid: uuid.UUID) -> None:
        for fc in (c for c in (item.get(['inference', 'classification']),) if c is not None and self.filter(c)):
            file, cf, lbl = fc.get('file', def_fl), fc['confidence'], fc['label']
            model = fc.get('ai_platform_model_id', fc.get('model_id', def_mdl))
            cs = list(fc['class_confidences'].values())

            self.append_entry(
                id=uuid.uuid4().bytes,
                timestamp=ts,
                frame_id=fid,
                file=file,
                model_id=model,
                confidence=cf,
                label=lbl,
                class_confidences=cs,
                geometry_type=GeometryType.TAG.value,
                geometry=[],
            )

    def append_roi_classification(
        self,
        item: Item,
        default_file: str,
        default_model_id: int,
        ts: datetime,
        fid: uuid.UUID,
    ) -> None:
        for fc in (c for c in (item.get(['inference', 'roi_classification']),) if c is not None and self.filter(c)):
            file, cf, lbl = fc.get('file', default_file), fc['confidence'], fc['label']
            model = fc.get('ai_platform_model_id', fc.get('model_id', default_model_id))
            cs = list(fc['class_confidences'].values())

            try:
                geometry = _get_roi_geometry(item, file, fc['roi'])
            except ValueError as e:
                logger.warning(f'Could not encode roi geometry: {e}')
                continue

            self.append_entry(
                id=uuid.uuid4().bytes,
                timestamp=ts,
                frame_id=fid,
                file=file,
                model_id=model,
                confidence=cf,
                label=lbl,
                class_confidence=cs,
                geometry_type=GeometryType.ROI2D.value,
                geometry=geometry,
            )


class TrackedObject(Table[SchemaTrackedObject]):
    """Table for tracked objects."""

    @classmethod
    def parents(cls) -> list[type[Table[T]]]:
        """Return the list of tables rows of this table depend on."""
        return [Frame]

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        timestamp = next(new_data[Frame.table_name()].item_data('timestamp'))
        for t in (t for t in item.get(['inference', 'tracking', 'tracked_objects'], []) if self.filter(t)):
            (x, y, *oz), (vx, vy, *ovz) = _get(t, 'position'), _get(t, 'velocity')
            z, vz = oz[0] if len(oz) else 0.0, ovz[0] if len(ovz) else 0.0

            self.append_entry(
                id=self.get_uuid(_get(t, 'id')).bytes, timestamp=timestamp, x=x, y=y, z=z, v_x=vx, v_y=vy, v_z=vz
            )


class DistinctTrackedObject(Table[SchemaTrackedObjectDistinct]):
    """Table for distinct tracked objects."""

    @classmethod
    def parents(cls) -> list[type[Table[T]]]:
        """Return the list of tables rows of this table depend on."""
        return [Frame]

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        timestamp = next(new_data[Frame.table_name()].item_data('timestamp'))
        for t in (t for t in item.get(['inference', 'tracking', 'added_tracked_objects'], []) if self.filter(t)):
            self.append_entry(id=self.get_uuid(_get(t, 'id')).bytes, timestamp=timestamp)


class Event(Table[SchemaEvent]):
    """Table for tracked objects."""

    @classmethod
    def parents(cls) -> list[type[Table[T]]]:
        """Return the list of tables rows of this table depend on."""
        return [Frame]

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        timestamp = next(new_data[Frame.table_name()].item_data('timestamp'))
        for e in (e for e in item.get(['inference', 'events'], []) if self.filter(e)):
            self.append_entry(
                id=uuid.uuid4().bytes,
                timestamp=timestamp,
                type=e['warning_type'],
                start_time=datetime.utcfromtimestamp(e['warning_start_time'] * 1e-3),
                referenced_entity_id=self.get_uuid(e['warning_id']).bytes,
                referenced_table='tracked_object',
            )


class DistinctEvent(Table[SchemaEventDistinct]):
    """Table for distinct tracked objects."""

    @classmethod
    def parents(cls) -> list[type[Table[T]]]:
        """Return the list of tables rows of this table depend on."""
        return [Frame, Event]

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        super().from_item(item, new_data)
        timestamp = next(new_data[Frame.table_name()].item_data('timestamp'))

        item_events: list[dict[str, Any]] = [e for e in item.get(['inference', 'events'], []) if self.filter(e)]
        e_ids: Iterator[bytes] = new_data[Event.table_name()].item_data('referenced_entity_id')

        for e, e_id in zip(item_events, e_ids, strict=True):
            if e.get('warning_is_new', False):
                self.append_entry(id=e_id, timestamp=timestamp, type=e['warning_type'])


def _get_roi_geometry(item: Item, file: str, roi_name: str) -> list[float]:
    if file not in item:
        raise ValueError(f'File "{file}" not in item even though there is a roi classification for {roi_name}.')
    x_left, x_right, y_top, y_bottom = get_region_of_interest(roi_name)
    ih, iw = item[file].shape[:2]
    w, h = (x_right - x_left) / float(iw), (y_bottom - y_top) / float(ih)
    x, y = (x_right + x_left) / float(iw) / 2.0, (y_bottom + y_top) / float(ih) / 2.0
    return [float(roi_pos(roi_name)), x, y, w, h, 0.0]
