"""Definition of tables for insights.

This module provides core abstractions for tables and table collections used
to organize, manage, and process insights.

 -  The `Table` class represents an individual table with a specific schema.
 -  The `TableCollection` class manages the hierarchy and relationships between tables using a
    Directed Acyclic Graph (DAG).
"""

import uuid
import copy
import functools

import polars as pl

import networkx

from logging import getLogger
from typing import TypeVar, Generic, get_origin, get_args, Any, Optional, Callable, Iterator, get_type_hints
from collections.abc import Iterable

from dataspree.inference.core.config import camel_to_snake
from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item

from dataspree.inference.insights.schemata import BaseSchema

T = TypeVar('T', bound=BaseSchema)

logger = getLogger(__name__)


class TableType(type):
    """A table's type.

    Keeps track of the implementations despite the use of generics.
    Also allows for nicer type hinting.
    """

    _subclasses: list['TableType'] = []

    def __new__(cls, name: str, bases: tuple[type], dct: dict[str, Any]) -> 'TableType':
        """Create a new subclass."""
        new_class = super().__new__(cls, name, bases, dct)
        try:
            Table._subclasses.append(new_class)
        except NameError:
            pass

        return new_class

    @classmethod
    def all_subclasses(cls) -> list['TableType']:
        """Return all subclasses of a particular class."""
        return cls._subclasses


class Table(Generic[T], metaclass=TableType):
    """Core Table abstraction.

    Represents a table with a specific schema and provides operations for managing
    rows, schema validation, and hierarchical relationships with parent tables.

    Attributes:
        __session_uuid (uuid.UUID): A unique session UUID used for generating
                                    deterministic UUIDs for table entries.
    """

    __session_uuid: uuid.UUID = uuid.uuid4()

    def __init__(self) -> None:
        """Initialize the table.

        Sets up the schema fields, an empty storage for table rows, and a default
        filter function for processing rows.
        """
        self._template_fields: dict[str, type] = dict(self.schema().__annotations__.items())
        self._fields: dict[str, Any] = {k: [] for k in self._template_fields.keys()}
        self.filter: Callable[[dict[Any, Any]], bool] = lambda _a: True
        self.__length: int = 0

    @property
    def data(self) -> dict[str, list[Any]]:
        return self._fields

    def clear(self) -> None:
        for v in self._fields.values():
            v.clear()

    @classmethod
    def table_name(cls) -> str:
        """Return the name of the table.

        Returns:
            str: The table's name in snake_case.
        """
        name: str = camel_to_snake(cls.__name__)
        return name

    def from_item(self, item: Item, new_data: dict[str, 'Table[BaseSchema]']) -> None:
        """Populate the table with data from an item.

        Args:
            item (Item): The source item.

            new_data (dict[str, Table[BaseSchema]): New data from other tables.
        """
        self.__length = len(next(iter(self._fields.values()), []))

    def item_data(self, column: str) -> Iterator[Any]:
        """Return the new data for a specific table name and column."""
        clm = self.data[column]
        return (clm[i] for i in range(self.__length, len(clm)))

    def append_entry(self, _append: bool = True, **kwargs: Any) -> None:
        """Add a new entry to the table.

        Args:
            _append (bool): Whether to append or extend the entries. Defaults to True.

            **kwargs: The fields for the entry.

        Raises:
            KeyError: If required fields are missing or extra fields are provided.
        """
        if __debug__ and is_application_debug():
            missing_fields = [k for k in self._template_fields if k not in kwargs.keys()]
            if len(missing_fields):
                raise KeyError(f'Missing argument(s) {", ".join(missing_fields)} in table "{self.table_name()}".')

            extra_fields = [k for k in kwargs.keys() if k not in self._template_fields]
            if len(extra_fields):
                raise KeyError(f'Extra argument(s) {", ".join(extra_fields)} in table "{self.table_name()}".')

            if not _append:
                lengths_set: set[int] = {len(v) if hasattr(v, '__len__') else 1 for v in kwargs.values()}

                if len(lengths_set) > 1:
                    raise ValueError(f'Inconsistent field lengths: {lengths_set} in table "{self.table_name()}".')

        for k, v in self._fields.items():
            if _append:
                v.append(kwargs[k])
            else:
                v.extend(kwargs[k])

    @classmethod
    def parents(cls) -> list[type['Table[BaseSchema]']]:
        """Return the list of tables this table depends on. Used for placement in the DAG.

        Returns:
            list[type[Table[BaseSchema]]]: A list of parent table types.
        """
        return []

    # Utilities

    @classmethod
    def get_uuid(cls, internal_id: int, entity_type_name: Optional[str] = None) -> uuid.UUID:
        """Utility method that generates a UUID based on the entity type and internal ID.

        Args:
            internal_id (int):
                The internal ID of the entity.

            entity_type_name (Optional[str]):
                The name of the entity type. Defaults to the table name.

        Returns:
            uuid.UUID: A UUID derived from the session UUID and entity details.
        """
        return uuid.uuid5(Table.__session_uuid, f'{entity_type_name or cls.table_name()}-{internal_id}')
        # return uuid.uuid4()

    # Helper functions to access the template argument.

    @classmethod
    def orig_bases(cls) -> tuple[type, ...]:
        """Retrieve and return the `__orig_bases__` of a class.

        Returns:
            list[type]: The original bases of the class, or an empty list if not available.
        """
        return getattr(cls, '__orig_bases__', tuple())

    @classmethod
    @functools.cache
    def polars_schema(cls) -> dict[str, Any]:
        """Return the polars schema of the table.

        Returns:
            dict[str, pl.DataType]: The schema type representing the table's structure.

        Raises:
            ValueError: If the schema cannot be resolved.
        """

        def _convert_list(type_hint: Any) -> Any:
            if get_origin(type_hint) is list:
                inner = get_args(type_hint)[0]
                return pl.List(inner)
            return type_hint

        schema = {k: _convert_list(v) for k, v in get_type_hints(cls.schema()).items()}
        return schema

    @classmethod
    @functools.cache
    def schema(cls):
        """Return the schema type (T) of the table.

        Returns:
            type[T]: The schema type representing the table's structure.

        Raises:
            ValueError: If the schema cannot be resolved.
        """
        for base in cls.orig_bases():
            origin = get_origin(base)

            while origin is not Table and origin is not None:
                base = origin
                origin = get_origin(base)

            if origin is None:
                continue

            assert origin is Table

            args_base: tuple[type[T], ...] = get_args(base)
            if len(args_base) != 1:
                raise ValueError('Expected exactly one generic argument to Data.')
            return args_base[0]

        raise ValueError(f'Base schema inaccessible for table {cls.__name__}.')


TTable = TypeVar('TTable', bound=Table[BaseSchema])


class TableCollection:
    """Manages a collection of tables and their dependencies.

    Provides dictionary-like access to tables and manages relationships between
    tables using a Directed Acyclic Graph (DAG).
    """

    def __init__(self) -> None:
        """Initialize the table collection.

        Sets up storage for tables, a topological sort of the tables, and the
        dependency DAG.
        """
        self._tables: dict[str, Table[BaseSchema]] = {}
        self._sorted_tables: list[Table[BaseSchema]] = []
        self.__dag = networkx.DiGraph()

    def register_table(self, table: type[TTable]) -> TTable:
        """Register a table and its hierarchy in the collection.

        Args:
            table (type[Table[BaseSchema]]): The table type to register.

        Returns:
            Table[BaseSchema]: The registered table instance.

        Raises:
            ValueError: If the table is already registered or introduces a cycle.
        """
        if table.table_name() in self._tables:
            raise ValueError(f'Table {table.table_name()} already registered.')

        # Attempt to build DAG
        _dag = copy.deepcopy(self.__dag)
        _dag.add_node(child_table_name := table.table_name())
        _table_types: dict[str, type[Table[BaseSchema]]] = {child_table_name: table}
        for parent in table.parents():
            parent_name = parent.table_name()
            _table_types[parent_name] = parent
            _dag.add_node(parent_name)
            _dag.add_edge(parent_name, child_table_name)

        # Validate if the graph is still a DAG
        if not networkx.is_directed_acyclic_graph(_dag):
            raise ValueError(f'Adding {child_table_name} introduces a cycle in the DAG.')

        # Use topological sorting.
        ts: list[str] = list(networkx.topological_sort(_dag))
        self._tables = {table_name: self._tables.get(table_name) or _table_types[table_name]() for table_name in ts}
        self._sorted_tables = [self._tables[table_name] for table_name in ts]
        self.__dag = _dag
        logger.info(f'Registered Table {child_table_name}: {", ".join(ts)}')

        return self._tables[child_table_name]

    def from_item(self, item: Item) -> None:
        """Populate all tables in the collection using a single item.

        Args:
            item (Item): The source item.
        """
        for table in self._sorted_tables:
            table.from_item(item, self._tables)

    def data(self) -> dict[str, dict[str, list[Any]]]:
        """Collect all data from the tables in the collection.

        Returns:
            dict[str, dict[str, list[Any]]]: Collected data organized by table
                and field name.
        """
        return {table.table_name(): table.data for table in self._sorted_tables}

    def clear(self) -> None:
        for st in self._sorted_tables:
            for v in st.data.values():
                v.clear()

    # Utilities for iterating over the tables.

    def __getitem__(self, key: str) -> Table[BaseSchema]:
        """Retrieve a table by its key.

        Args:
            key (str): The table name.

        Returns:
            Table[BaseSchema]: The table corresponding to the key.
        """
        return self._tables[key]

    def __iter__(self) -> Iterable[str]:
        """Iterate over the keys of the collection."""
        return iter(self._tables)

    def __len__(self) -> int:
        """Return the number of tables in the collection.

        Returns:
            int: The number of tables.
        """
        return len(self._tables)

    def __contains__(self, key: str) -> bool:
        """Check if a table exists in the collection.

        Args:
            key (str): The table name.

        Returns:
            bool: True if the table exists, False otherwise.
        """
        return key in self._tables

    def items(self) -> Iterable[tuple[str, Table[BaseSchema]]]:
        """Return a view of the items in the collection.

        Returns:
            Iterable[tuple[str, Table[BaseSchema]]]: An iterable of key-value pairs.
        """
        return self._tables.items()

    def keys(self) -> Iterable[str]:
        """Return a view of the keys in the collection.

        Returns:
            Iterable[str]: An iterable of table names.
        """
        return self._tables.keys()

    def values(self) -> Iterable[Table[BaseSchema]]:
        """Return a view of the values in the collection.

        Returns:
            Iterable[Table[BaseSchema]]: An iterable of table objects.
        """
        return self._tables.values()
